#include <QApplication>
#include <QScreen>
#include <QRect>
#include <QList>
#include <QWindow>
#include <QWidget>
#include <QVBoxLayout>
#include <QTimer>
#include <QLabel>

#include <filesystem>
#include <fstream>
#include <vector>
#include <optional>
#include <exception>
#include <iostream>
#include <format>

#include "config.h"

class Dock {
    private:
        /// @brief Stores all dock windows    
        QList<QWindow*> m_dockWindows;

        /// @brief QWidget container for dock windows
        QList<QWidget*> m_dockWindowContainters;

        /// @brief Stores all configuration windows
        QList<QWindow*> m_configWindows;

        /// @brief Stores all available screens
        QList<QScreen*> m_screens = QGuiApplication::screens();

        /// @brief An application in the dock
        struct dock_application {
            /// @brief Whether the application is pinned to the dock
            bool pinned;

            /// @brief Index of where the application is pinned (0 when disabled)
            int pinIndex;

            /// @brief Whether the application is running
            bool running;

            /// @brief Path to the application's desktop file entry
            std::filesystem::path desktopFilePath;

            /// @brief Reference the to the icon file of the application
            std::ifstream desktopIcon;
            
            /// @brief Name of the application as specified by the desktop file 
            std::string desktopName;

            /// @brief Name of the running process, used if an application desktop file does not exist
            std::string processName;

            /// @brief Create a new `dock_application` with an optional desktop file
            /// @param filename Optional desktop file for metadata
            dock_application() {
                getDesktopName();
                getDesktopIcon();
            }
            
            /// @brief Clean up `dock_application` resources
            ~dock_application() {
                if (desktopIcon.is_open()) {
                    desktopIcon.close();
                }
            }

            /// @brief Get application desktop name from desktop file
            /// @return UTF-16 desktop name
            std::optional<std::string> getDesktopName() {
                inih::INIReader r;
                std::string name;

                try {
                    r = std::move(inih::INIReader(desktopFilePath.string()));
                } catch (std::runtime_error&) {
                    return std::nullopt;
                }

                try {
                    name = r.Get<std::string>("Desktop Entry", "name");
                } catch (std::runtime_error&) {
                    return std::nullopt;
                }

                return name;
            }

            /// @brief Get application desktop icon from desktop file
            /// @return Desktop icon file stream
            std::optional<std::ifstream> getDesktopIcon() {
                inih::INIReader r;
                std::filesystem::path icon_path;
                std::ifstream icon{icon_path};

                try {
                    r = std::move(inih::INIReader(desktopFilePath.string()));
                } catch (std::runtime_error&) {
                    return std::nullopt;
                }

                try {
                    icon_path = r.Get<std::string>("Desktop Entry", "icon");
                } catch (std::runtime_error&) {
                    return std::nullopt;
                }

                return icon;
            }
        };

        /// @brief List of all applications in the dock
        std::vector<dock_application> m_dockApplications;

        /// @brief Create dock windows on all screens
        /// @param app Qt Application callback
        void createDockWindows(QApplication* app);
    public:
        Dock(QApplication* app);
        ~Dock(); 
};